﻿//  *********************************************************
//  Configuration Section - Begin
//  *********************************************************

"use strict";
var mongoConnectorVersionStr = "5.4.6.17122101";
var mongoConnectorVersionNum = 50406.17122101;
var mongoStatus = "Running";
var connectorPort = process.env.APPX_MONGO_CONNECTOR_PORT;

var sslEnabled = false;
var cryptoEnabled = true;
var sslPrivateKey = "/etc/pki/tls/private/appx.key";
var sslCertificate = "/etc/pki/tls/certs/appx.com.crt";
var sslCertAuthority = "/etc/pki/tls/certs/gd_bundle.crt";

var mongoDatabase = "AppxDatabaseCache";
var mongoHost = "localhost";
var mongoPort = 27017;

//  *********************************************************
//  Configuration Section - End
//  *********************************************************

/************************************************************
    Library Imports - Begin
*************************************************************/
var StringLib = require('string');
var GridFSBucket = require('mongodb').GridFSBucket;
var MongoClient = require('mongodb').MongoClient;
var MongoServer = require('mongodb').Server;
var fs = require('fs');
var WebSocket = require('ws').WebSocket;
var WebSocketServer = require('ws').Server;
var net = require('net');
var hexy = require('hexy');
var StringDecoder = require('string_decoder').StringDecoder;
var decoder = new StringDecoder('utf8');
var crypto = require('crypto');
var node_cryptojs = require('node-cryptojs-aes');
/************************************************************
    Library Imports - End
*************************************************************/

// Global variabls and Objects

var mongoClient = new MongoClient(new MongoServer(mongoHost, mongoPort));
var mongoUrl = 'mongodb://' + mongoHost + ':' + mongoPort + '/' + mongoDatabase + '?maxPoolSize=20';
var mongoCacheDb = null;

mongoClient.connect(mongoUrl, function mongoClient_connectCallback(err, db) {
    mongoCacheDb = db;
    if (err) {
        mongoStatus = "Error";
    } else {
        createWebSocket();
    }
});

var buildMongoSelection = function buildMongoSelection(gridRequest) {
    var selection = [];
    var result = {};
    if (gridRequest["_search"] == "true") {
        var filters = JSON.parse(gridRequest.filters);
        var groupOp = "$" + filters.groupOp.toLowerCase();
        var rules = filters.rules;
        
        for (var i = 0; i < rules.length; i++) {
            var rule = rules[i];
            var field = rule.field;
            var op = rule.op;
            var data = rule.data;
            var mongo = {};
            var datatype = field.slice(-1);

            /*Change search fields back to correct fields, not sorting fields*/
            var colModelPosition = parseInt(field.substring(field.length - 2, field.length - 1));
            colModelPosition += colModelPosition - 3;
            var gridRequestPropertyIndex = "colModel[" + colModelPosition + "][index]";
            var gridRequestPropertyName = "colModel[" + colModelPosition + "][name]";
            if (field === gridRequest[gridRequestPropertyIndex]) {
                field = gridRequest[gridRequestPropertyName];
            }
            if (datatype === "D") {
                for (var key in gridRequest) {
                    if (gridRequest[key] === field && key.indexOf("index") !== -1) {
                        var tempKey = key.replace(/index/, "name");
                        if (gridRequest[tempKey] !== field) {
                            field = gridRequest[tempKey];
                        }
                    }
                }
            }


            switch (datatype) {
                case "I": data = parseInt(data); break;
                case "F": data = parseFloat(data); break;
                case "B": data == "Y"; break;
            }

            mongo[field] = {};

            /* Test for uppercase letters in search string. If search string
            ** contains uppercase then we do case sensitive search. Otherwise
            ** searches are case insensitive.*/
            if (/[A-Z]/.test(data) || !isNaN(data)) {
                switch (op) {
                    case 'eq': mongo[field]["$eq"] = data; break;
                    case 'ne': mongo[field]["$ne"] = data; break;
                    case 'lt': mongo[field]["$lt"] = data; break;
                    case 'gt': mongo[field]["$gt"] = data; break;
                    case 'le': mongo[field]["$lte"] = data; break;
                    case 'ge': mongo[field]["$gte"] = data; break;
                    case 'nu': mongo[field]["$eq"] = " "; break;
                    case 'nn': mongo[field]["$ne"] = " "; break;
                    case 'bw': mongo[field]["$regex"] = "^" + data; break;
                    case 'ew': mongo[field]["$regex"] = data + "$"; break;
                    case 'cn': mongo[field]["$regex"] = data; break;
                    case 'bn': mongo[field]["$regex"] = "^(?!" + data + ").*"; break;
                    case 'en': mongo[field]["$regex"] = ".*(?<!" + data + ")$"; break;
                    case 'nc': mongo[field]["$regex"] = "^((?!" + data + ").)*$"; break;
                }
            } else {
                switch (op) {
                    case 'eq': mongo[field]["$regex"] = "^(" + data + ")$"; break;
                    case 'ne': mongo[field]["$regex"] = "^(?!" + data + ")$"; break;
                    case 'lt': mongo[field]["$lt"] = data; break;
                    case 'gt': mongo[field]["$gt"] = data; break;
                    case 'le': mongo[field]["$lte"] = data; break;
                    case 'ge': mongo[field]["$gte"] = data; break;
                    case 'nu': mongo[field]["$eq"] = " "; break;
                    case 'nn': mongo[field]["$ne"] = " "; break;
                    case 'bw': mongo[field]["$regex"] = "^" + data; break;
                    case 'ew': mongo[field]["$regex"] = data + "$"; break;
                    case 'cn': mongo[field]["$regex"] = data; break;
                    case 'bn': mongo[field]["$regex"] = "^(?!" + data + ").*"; break;
                    case 'en': mongo[field]["$regex"] = ".*(?<!" + data + ")$"; break;
                    case 'nc': mongo[field]["$regex"] = "^((?!" + data + ").)*$"; break;
                }
            }

            if (mongo[field].hasOwnProperty("$regex")) {
                mongo[field]["$options"] = "i";
            }
            selection.push(mongo);
        }
        result[groupOp] = selection;
    }
    return result;
}

var buildMongoSort = function buildMongoSort(gridRequest) {
    var sorter = {};
    var ord = 1;

    if (gridRequest.sidx && gridRequest.sidx.length > 0) {
        if (gridRequest.sord && gridRequest.sord == 'asc')
            sorter[gridRequest.sidx] = 1;
        else {
            sorter[gridRequest.sidx] = -1;
            ord = -1;
        }
    }
    /*If table was previously sorted we use that as a secondary sort*/
    if (Object.hasOwnProperty.call(gridRequest, "lastSortName[]") && gridRequest["lastSortName[]"].length >= 1) {
        for (var i = gridRequest["lastSortName[]"].length - 1; i >= 0; i--) {
            if (gridRequest["lastSortName[]"][i] !== gridRequest.sidx) {
                if (gridRequest["lastSortOrder[]"][i] === "asc") {
                    sorter[gridRequest["lastSortName[]"][i]] = 1;
                } else {
                    sorter[gridRequest["lastSortName[]"][i]] = -1;
                }
            }
        }
    } else {
        sorter["initialSort"] = ord;
    }

    return sorter;
}

var fetchtabledata = function fetchtabledata(message, res) {
    // This code is full of nested callbacks.  We have to open the database, then 
    // in that callback we have to open the collection, then inside that callback we have to
    // process the count function, then in that callback we get the requested records, and inside 
    // that callback we send the records to the client.  Whew, that's a lot of callbacks.

    // We are in the OPEN callback
    try {
        mongoCacheDb.collection(message.collection, function mongoCacheDb_collectionCallback(err, collection) {
            // We are in the GET COLLECTION callback
            var selection = buildMongoSelection(message);
            var sorter = buildMongoSort(message);
            var index = {};
            var collist = {};

            if (message.collist) {
                collist = JSON.parse(message.collist);
            }

            collection.createIndex(sorter);
            collection.find({ $and: [{ "datalookupid": message.datalookupid }, selection] }).count(function mongoCountCallback(err, totalcount) {
                // We are in the COUNT results callback 
                collection.find({ $and: [{ "datalookupid": message.datalookupid }, selection] }, collist).sort(sorter).skip((parseInt(message.page) - 1) * parseInt(message.rows)).limit(parseInt(message.rows)).toArray(function mongoFindCallback(err, items) {
                    // We are in the GET RECORDS callback
                    var result = {};
                    result.records = parseInt(totalcount);
                    result.page = parseInt(message.page);
                    result.total = Math.ceil(parseInt(totalcount) / parseInt(message.rows));
                    var rowtmp = [];
                    if (items) {
                        for (var i = 0; i < items.length; i++) {
                            rowtmp.push(items[i]);
                        }
                    }
                    result["rows"] = rowtmp;
                    res.end(JSON.stringify(result));
                });
            });
        });
    } catch (e) {
        console.log("ERROR: connecting to db");
        console.log(e);
        console.log(e.stack);
    }
};


function createWebSocket() {
    var CryptoJS = node_cryptojs.CryptoJS;
    var JsonFormatter = node_cryptojs.JsonFormatter;
    var fileBuffer = Buffer.alloc(0);
    if (sslEnabled)
        var https = require('https');
    else
        var https = require('http');

    function processDataRequest(req, res) {
        var postData = "";
        var query;
        var url = require('url');
        var url_parts = url.parse(req.url, true);
        var stream = require('streamifier');
        if (req.method === "POST") {
            if (url_parts.pathname.indexOf("/upload/") != -1) {
                /*If we are sending file to be uploaded to mongo*/
                    var fgb = url_parts.pathname.substring(url_parts.pathname.indexOf("/upload/") + 8);
                    fgb = fgb.substring(0, fgb.indexOf("/"));
                    var skipLen = fgb.length + 9;
                    var gb = new GridFSBucket(mongoCacheDb, { bucketName: fgb });
                    var fileName = url_parts.pathname.substring(url_parts.pathname.indexOf("/upload/") + skipLen);
                    var uploadStream = gb.openUploadStream(fileName);
                    uploadStream.on("error", function uploadStream_onError(error) {
                        res.write("Error uploading file to mongo database");
                        console.log(error);
                        uploadStream = null;
                    });
                    uploadStream.once("finish", function uploadStream_onceFinish() {
                        res.writeHead("201", { 'Access-Control-Allow-Origin': '*', 'Bucket': fgb });
                        res.end();
                    });
                    req.pipe(uploadStream);
            }
            req.on("data", function req_onDataCallback(chunk) {
                if (url_parts.pathname.indexOf("/upload/") === -1) {
                    postData += chunk.toString();
                }
            });
            req.on("end", function req_onEndCallback() {
                if (url_parts.pathname.indexOf("/upload/") === -1) {
                    var qString = require("querystring");
                    query = qString.parse(postData);
                    getData();
                } else {
                    
                }
            });
        } else {
            query = url_parts.query;
            getData();
        }

        function getData() {
            var mime = require('mime');
            /*
            **Function send HTTP error back to browser for display.
            **
            **@param mRes: Response to http request
            */
            function fileNotFound(mRes) {
                /*If file is not found, then we return a file not found
                **error*/
                res.writeHead(404, { 'Content-Type': 'text/plain' });
                res.write('404 Not Found\n');
                res.end();
            }
            if (url_parts.pathname.indexOf('/getGridData') != -1) {
                res.writeHead(200, { 'Access-Control-Allow-Origin': '*', 'Content-Type': 'application/json' });
                fetchtabledata(query, res);
            }
            else if (url_parts.pathname.indexOf("/getResource/") != -1) {
                var fileData = Buffer.alloc(0);
                gb = new GridFSBucket(mongoCacheDb, { bucketName: "resource" });
                var fileName = encodeURI(url_parts.pathname.substring(url_parts.pathname.indexOf("/getResource/") + 13));

                var downloadStream = gb.openDownloadStreamByName(fileName);
                downloadStream.on("error", function downloadStream_onError(error) {
                    fileNotFound(res);
                });
                downloadStream.on("data", function downloadStream_onData(data) {
                    var totBufLen = fileData.length + data.length;
                    fileData = Buffer.concat([fileData, data], totBufLen);
                });
                downloadStream.on("end", function downloadStream_onEnd() {
                    var mimeType = mime.getType(fileName);
                    res.writeHead(200, { 'Access-Control-Allow-Origin': '*', 'Content-Type': mimeType, 'Cache-Control': 'max-age=31536000' });
                    stream.createReadStream(fileData).pipe(res);
                });
            }
            else if (url_parts.pathname.indexOf("/getFile/") != -1) {
                /*If we pushed file into mongo to be displayed in the browser or use
                **the browser to save file to client*/
                var fileData = Buffer.alloc(0);
                var fgb = url_parts.pathname.substring(url_parts.pathname.indexOf("/getFile/") + 9);
                fgb = fgb.substring(0, fgb.indexOf("/"));
                var skipLen = fgb.length + 10;
                var gb = new GridFSBucket(mongoCacheDb, { bucketName: fgb });

                var fileName = url_parts.pathname.substring(url_parts.pathname.indexOf("/getFile/") + skipLen);

                var downloadStream = gb.openDownloadStreamByName(fileName);
                downloadStream.on("error", function downloadStram_onError(error) {
                    fileNotFound(res);
                });
                downloadStream.on("data", function downloadStream_onData(data) {
                    var totBufLen = fileData.length + data.length;
                    fileData = Buffer.concat([fileData, data], totBufLen);
                });
                downloadStream.on("end", function downloadStream_onEnd() {
                    if (this.s.file === null) {
                        fileNotFound(res);
                        return;
                    }
                    var id = this.s.file._id;
                    var mimeType = mime.getType(fileName);
                    res.writeHead(200, { 'Access-Control-Allow-Origin': '*', 'Content-Type': mimeType, 'Cache-Control': 'no-cache, must-revalidate' });
                    stream.createReadStream(fileData).pipe(res);
                    gb.delete(id);
                });
            }
            else if (url_parts.pathname.indexOf("/sendFile/") != -1) {
                /*If we pushed file into mongo to be uploaded to appx*/
                var fileData = Buffer.alloc(0);
                var fgb = url_parts.pathname.substring(url_parts.pathname.indexOf("/sendFile/") + 10);
                fgb = fgb.substring(0, fgb.indexOf("/"));
                var skipLen = fgb.length + 11;
                var gb = new GridFSBucket(mongoCacheDb, { bucketName: fgb });

                var fileName = encodeURI(url_parts.pathname.substring(url_parts.pathname.indexOf("/sendFile/") + skipLen));

                var downloadStream = gb.openDownloadStreamByName(fileName);
                downloadStream.on("error", function downloadStram_onError(error) {
                    fileNotFound(res);
                });
                downloadStream.on("data", function downloadStream_onData(data) {
                    var totBufLen = fileData.length + data.length;
                    fileData = Buffer.concat([fileData, data], totBufLen);
                });
                downloadStream.on("end", function downloadStream_onEnd() {
                    var mimeType = mime.getType(fileName);
                    res.writeHead(200, { 'Access-Control-Allow-Origin': '*', 'Content-Type': mimeType, 'Cache-Control': 'no-cache, must-revalidate' });
                    stream.createReadStream(fileData).pipe(res);
                });
            }
        }
    }

    // Configure the websocket
    if (sslEnabled) {
        var options = {
            key: fs.readFileSync(sslPrivateKey),
            cert: fs.readFileSync(sslCertificate),
            ca: fs.readFileSync(sslCertAuthority)
        };

        var app = https.createServer(options, function https_createServerCallback(req, res) {
            processDataRequest(req, res);
        }).listen(connectorPort);
    }
    else {
        var app = https.createServer(function https_createServerCallback(req, res) {
            processDataRequest(req, res);
        }).listen(connectorPort);
    }

    // start a websocket server
    var wss = new WebSocketServer({
        server: app
    });
}

